@@ -22,7 +22,7 @@ def bs(byte_array):
     return ''.join([chr(c) for c in byte_array])
 
 DEBUG = False
-__version__ = '0.3.5'
+__version__ = '0.4.0'
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'qmark'
@@ -798,7 +798,7 @@ class BaseConnect:
         p.pack_int(self.op_attach)
         p.pack_int(2)   # CONNECT_VERSION2
         p.pack_int(1)   # Arch type (Generic = 1)
-        p.pack_string(self.str_to_bytes(self.filename))
+        p.pack_string(self.str_to_bytes(self.filename if self.filename else ''))
         p.pack_int(1)   # Protocol version understood count.
         p.pack_bytes(self.uid())
         p.pack_int(10)  # PROTOCOL_VERSION10
@@ -1157,14 +1157,19 @@ class BaseConnect:
         (h, oid, buf) = self._op_response()
         delattr(self, "db_handle")
 
-    def __init__(self, dsn, user, password, charset=DEFAULT_CHARSET, port=3050):
-        i = dsn.find(':')
-        if i < 0:
-            self.hostname = ''
-            self.filename = dsn
+    def __init__(self, dsn=None, user=None, password=None, host=None,
+            database=None, charset=DEFAULT_CHARSET, port=3050):
+        if dsn:
+            i = dsn.find(':')
+            if i < 0:
+                self.hostname = host
+                self.filename = dsn
+            else:
+                self.hostname = dsn[:i]
+                self.filename = dsn[i+1:]
         else:
-            self.hostname = dsn[:i]
-            self.filename = dsn[i+1:]
+            self.hostname = host
+            self.filename = database
         self.user = user
         self.password = password
         self.charset = charset
@@ -1203,8 +1208,10 @@ class BaseConnect:
             self.close()
 
 class connect(BaseConnect):
-    def __init__(self, dsn, user, password, charset=DEFAULT_CHARSET , port=3050):
-        BaseConnect.__init__(self, dsn, user, password, charset, port)
+    def __init__(self, dsn=None, user=None, password=None, host=None,
+            database=None, charset=DEFAULT_CHARSET, port=3050):
+        BaseConnect.__init__(self, dsn=dsn, user=user, password=password,
+                    host=host, database=database, charset=charset, port=port)
         self._op_connect()
         self._op_accept()
         self._op_attach()
@@ -1213,10 +1220,11 @@ class connect(BaseConnect):
 
 
 class create_database(BaseConnect):
-    def __init__(self, dsn, user, password, charset=DEFAULT_CHARSET , port=3050,
-                                                            page_size = 4096):
+    def __init__(self, dsn=None, user=None, password=None, host=None,
+            database=None, charset=DEFAULT_CHARSET, port=3050, page_size=4096):
         self.page_size = page_size
-        BaseConnect.__init__(self, dsn, user, password, charset, port)
+        BaseConnect.__init__(self, dsn=dsn, user=user, password=password,
+                    host=host, database=database, charset=charset, port=port)
         self._op_connect()
         self._op_accept()
         self._op_create(self.page_size)
@@ -1224,21 +1232,23 @@ class create_database(BaseConnect):
         self.db_handle = h
 
 class service_mgr(BaseConnect):
-    def __init__(self, dsn, user, password, charset=DEFAULT_CHARSET , port=3050):
-        BaseConnect.__init__(self, dsn, user, password, charset, port)
+    def __init__(self, dsn=None, user=None, password=None, host=None,
+            database=None, charset=DEFAULT_CHARSET, port=3050):
+        BaseConnect.__init__(self, dsn=dsn, user=user, password=password,
+                    host=host, database=database, charset=charset, port=port)
         self._op_connect()
         self._op_accept()
         self._op_service_attach()
         (h, oid, buf) = self._op_response()
         self.db_handle = h
 
-    def backup_database(self, backup_filename, f=None):
+    def backup_database(self, backup_filename, callback=None):
         spb = bs([isc_action_svc_backup])
         s = self.str_to_bytes(self.filename)
         spb += bs([isc_spb_dbname]) + int_to_bytes(len(s), 2) + s
         s = self.str_to_bytes(backup_filename)
         spb += bs([isc_spb_bkp_file]) + int_to_bytes(len(s), 2) + s
-        if f:
+        if callback:
             spb += bs([isc_spb_verbose])
         self._op_service_start(spb)
         (h, oid, buf) = self._op_response()
@@ -1249,15 +1259,16 @@ class service_mgr(BaseConnect):
             if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
                 break
             ln = bytes_to_int(buf[1:2])
-            (f if f else sys.stdout).write(self.bytes_to_str(buf[3:3+ln]))
+            if callback:
+                callback(self.bytes_to_str(buf[3:3+ln]))
 
-    def restore_database(self, restore_filename, f=None):
+    def restore_database(self, restore_filename, callback=None):
         spb = bs([isc_action_svc_restore])
         s = self.str_to_bytes(restore_filename)
         spb += bs([isc_spb_bkp_file]) + int_to_bytes(len(s), 2) + s
         s = self.str_to_bytes(self.filename)
         spb += bs([isc_spb_dbname]) + int_to_bytes(len(s), 2) + s
-        if f:
+        if callback:
             spb += bs([isc_spb_verbose])
         spb += bs([isc_spb_res_buffers,0x00,0x08,0x00,0x00,isc_spb_res_page_size,0x00,0x10,0x00,0x00,isc_spb_options,0x00,0x30,0x00,0x00])
         self._op_service_start(spb)
@@ -1269,7 +1280,84 @@ class service_mgr(BaseConnect):
             if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
                 break
             ln = bytes_to_int(buf[1:2])
-            (f if f else sys.stdout).write(self.bytes_to_str(buf[3:3+ln]))
+            if callback:
+                callback(self.bytes_to_str(buf[3:3+ln]))
+
+    def trace_start(self, name=None, cfg=None, callback=None):
+        spb = bs([isc_action_svc_trace_start])
+        if name:
+            s = self.str_to_bytes(name)
+            spb += bs([isc_spb_trc_name]) + int_to_bytes(len(s), 2) + s
+        if cfg:
+            s = self.str_to_bytes(cfg)
+            spb += bs([isc_spb_trc_cfg]) + int_to_bytes(len(s), 2) + s
+        self._op_service_start(spb)
+        (h, oid, buf) = self._op_response()
+        self.svc_handle = h
+        while True:
+            self._op_service_info(bs([0x02]), bs([0x3e]))
+            (h, oid, buf) = self._op_response()
+            if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
+                break
+            ln = bytes_to_int(buf[1:2])
+            if callback:
+                callback(self.bytes_to_str(buf[3:3+ln]))
+
+    def trace_stop(self, id, callback=None):
+        id = int(id)
+        spb = bs([isc_action_svc_trace_stop])
+        spb += bs([isc_spb_trc_id]) + int_to_bytes(id, 4)
+        self._op_service_start(spb)
+        (h, oid, buf) = self._op_response()
+        self.svc_handle = h
+
+        self._op_service_info(bs([0x02]), bs([0x3e]))
+        (h, oid, buf) = self._op_response()
+        ln = bytes_to_int(buf[1:2])
+        if callback:
+            callback(self.bytes_to_str(buf[3:3+ln]))
+
+    def trace_suspend(self, id, callback=None):
+        id = int(id)
+        spb = bs([isc_action_svc_trace_suspend])
+        spb += bs([isc_spb_trc_id]) + int_to_bytes(id, 4)
+        self._op_service_start(spb)
+        (h, oid, buf) = self._op_response()
+        self.svc_handle = h
+
+        self._op_service_info(bs([0x02]), bs([0x3e]))
+        (h, oid, buf) = self._op_response()
+        ln = bytes_to_int(buf[1:2])
+        if callback:
+            callback(self.bytes_to_str(buf[3:3+ln]))
+
+    def trace_resume(self, id, callback=None):
+        id = int(id)
+        spb = bs([isc_action_svc_trace_resume])
+        spb += bs([isc_spb_trc_id]) + int_to_bytes(id, 4)
+        self._op_service_start(spb)
+        (h, oid, buf) = self._op_response()
+        self.svc_handle = h
+
+        self._op_service_info(bs([0x02]), bs([0x3e]))
+        (h, oid, buf) = self._op_response()
+        ln = bytes_to_int(buf[1:2])
+        if callback:
+            callback(self.bytes_to_str(buf[3:3+ln]))
+
+    def trace_list(self, callback=None):
+        spb = bs([isc_action_svc_trace_list])
+        self._op_service_start(spb)
+        (h, oid, buf) = self._op_response()
+        self.svc_handle = h
+        while True:
+            self._op_service_info(bs([0x02]), bs([0x3e]))
+            (h, oid, buf) = self._op_response()
+            if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
+                break
+            ln = bytes_to_int(buf[1:2])
+            if callback:
+                callback(self.bytes_to_str(buf[3:3+ln]))
 
     def close(self):
         if not hasattr(self, "db_handle"):
